package altro.reparto_ospedaliero.con_starvation;

import java.util.Vector;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

public class Stanza {

	private boolean giorno;

	// strutture di sincronizzazione
	private Vector<String> visite;// sostituisce "int visitatori;" ed è usato per comodità di stampa
	private Vector<String> medici; // sostituisce "boolean medico;" ed è usato per comodità di stampa

	// lock e condition:
	// diverse soluzioni alternative: ReentrantReadWriteLock, implementazione sul modello del problema Lettori/Scrittore, ...
	private Lock lock = new ReentrantLock();
	private Condition conditionVisitatori = lock.newCondition();
	private Condition conditionMedici = lock.newCondition();
	private Condition conditionDisplay = lock.newCondition();

	public Stanza() {
		giorno = true;
		// visitatori=0;
		visite = new Vector<String>();
		// medico = false;
		medici = new Vector<String>();
	}

	public void visitaMedica(String name) {

		lock.lock();
        // un medico è costretto ad attendere di giorno se è già in corso una visita medica nella stanza oppure se ci sono dei visitatori dentro
		while (giorno && (medici.size() > 0 || visite.size() > 0))
			try {
				conditionMedici.await();
			} catch (InterruptedException e) {
			}
		// medico = true;
		medici.add(name); //di notte potrebbero esserci più medici nella stessa stanza (il display li visualizzerà)
		conditionDisplay.signal();
		try {
		} finally {
			lock.unlock();
		}
	}

	public void terminaVisitaMedica(String name) {

		lock.lock();

		// medico = false;
		medici.remove(name);//di notte potrebbero esserci più medici nella stessa stanza (il display li visualizzerà)

		conditionMedici.signalAll(); // segnala a medici in attesa (di giorno) che la stanza è libera
		conditionVisitatori.signalAll();// segnala a visitatori in attesa 
		conditionDisplay.signal();

		try {
		} finally {
			lock.unlock();
		}
	}

	public void visitaCortesia(String name) {

		lock.lock();
       // un visitatore che vuol far visita al paziente di giorno deve attendere che termini una eventuale visita medica in corso, di notte deve accertarsi che non ci sia già un altro visitatore nella stanza 
		while ((giorno && medici.size() > 0) || (!giorno && visite.size() > 0))
			try {
				conditionVisitatori.await();
			} catch (InterruptedException e) {
			}

		//visitatori ++;
		visite.add(name); //potrebbero esserci più visitatori nella stessa stanza (il display li visualizzerà)
		conditionDisplay.signal();
		try {
		} finally {
			lock.unlock();
		}
	}

	public void terminaVisitaCortesia(String name) {

		lock.lock();
		//visitatori --;
		visite.remove(name);

		if (visite.size() == 0) {
			conditionMedici.signalAll(); //se la stanza si è "svuotata" dei visitatori un medico tra quelli in attesa potrebbe visitare il paziente
		}
		conditionDisplay.signal();

		try {
		} finally {
			lock.unlock();
		}
	}

	public void print() {

		try {
			lock.lock();

			while (true) {
				conditionDisplay.await();
				
				System.out.print(giorno?"\n@ | ":"\n( | ");

				for(String visitatore:visite)
					System.out.print(visitatore+ " ");

				for(String medico:medici)
					System.out.print(medico+ " ");
	
				System.out.println("|");
							}
		} catch (InterruptedException e) {
		} finally {
			lock.unlock();
		}
	}

	public void setRegime(boolean b) {
		giorno = b;
	}
}
